import numpy as np
import pandas as pd


def counts_to_csv(list_count_arr, graph, path_to_csv, sep=';', timesteps=None, use_index_as_col_name=False):
    """
    Take a list of numpy array generated by the count methods of the agent class, and save them as a CSV (easier for
    R users).

    :param list_count_arr: list of count arrays
    :param graph: graph object from which the count arrays have been extracted
    :param path_to_csv: full path to the final csv file
    :param sep: optional, string, default ';'. Separator in final CSV.
    :param timesteps: optional, list or array, default None. If not None, used to create a column 'timesteps' in the
                      csv.
    :param use_index_as_col_name: optional, boolean, default False. If True, the names of the columns in the CSV will
                                  be the index (i.e. integers) of the vertices of the graph. Else, it will be their
                                  id (generally a string or a couple).
    """
    row_stack = np.row_stack(list_count_arr)
    if use_index_as_col_name:
        columns = [str(i) for i in range(graph.connections.shape[0])]
    else:
        dict_ind_to_id = {ind: ids for ids, ind in graph.dict_cell_id_to_ind.items()}
        columns = [str(dict_ind_to_id[i]) for i in range(graph.connections.shape[0])]
    df = pd.DataFrame(row_stack, columns=columns)
    if timesteps is not None:
        df['timesteps'] = timesteps
    else:
        df['timesteps'] = np.array(range(1, row_stack.shape[0] + 1), dtype=np.int32)
    df.to_csv(path_to_csv, index=False, sep=sep)


def columns_to_csv(list_columns, list_col_names, path_to_csv, sep=';'):
    """
    Take a list of columns and a list of names to create a CSV.
    Columns are assumed to be the same length, and they should be iterable
    that Pandas can turn into a column of a Pandas.DataFrame object.

    :param list_columns: list whose element are colums of the CSV. That are, iterable
                         that pandas can cast into a Series and included in a Dataframe.
                         Each column should have the same length.
    :param list_col_names: list of string, name of each column. 
    """
    df = pd.DataFrame()
    for col, name in zip(list_columns, list_col_names):
        df[name] = col
    df.to_csv(path_to_csv, index=False, sep=sep)
    